/*
 * -----------------------------------------------------------------
 *    RICOH THETA Remote Shutter Client (Process) for PQI Air Pen
 * -----------------------------------------------------------------
 * mips-linux-gnu-gcc ThetaShutterProc.c --static -o mips/ThetaShutterProc
 *
 *  RICOH THETA HACKS! - MobileHackerz
 *  http://mobilehackerz.jp/contents/Review/RICOH_THETA
 *
 * This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://www.wtfpl.net/txt/copying/ for more details.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

const char g_sClientName[]  = "THETA_Remote";
const char g_sClientGUID[]  = {	0x39, 0x08, 0x93, 0xea, 0x8d, 0xd7, 0x44, 0xff, 0x95, 0xc7, 0x27, 0xa2, 0x20, 0xd5, 0x13, 0x36 };
/* 390893ea-8dd7-44ff-95c7-27a220d51336 */
const char g_AirPenLED[]    = "/proc/gpio/gpio23_out";

#define PTPIP_PACK_START                   (8)

/* PTP codes (named by libgphoto2) */
#define PTP_OC_OpenSession                 (0x1002)
#define PTP_OC_CloseSession                (0x1003)
#define PTP_OC_InitiateCapture             (0x100E)
#define PTP_OC_SetDevicePropValue          (0x1016)
#define PTP_EC_ObjectAdded                 (0x4002)
#define PTP_EC_CaptureComplete             (0x400D)
#define PTP_DPC_ExposureBiasCompensation   (0x5010)

/* prototypes */
int PTPIP_Init_Command_Request(int sock);
int PTPIP_Wait_Init_Command_Ack(int sock,int *session_id);
int PTPIP_Init_Event_Request(int sock, int session_id);
int PTPIP_Wait_Init_Event_Ack(int sock);
int PTPIP_Cmd_Request(int sock, int transaction_id, int cmd, int arg[], int arglen, char *payload, int payloadlen);
int PTPIP_Wait_Cmd_Response(int sock, char **data, int *data_len);
int PTPIP_Open_Connection(char *addr, int port);
int PTPIP_Close_Connection(int sock);
long PTPIP_Get_DWORD(unsigned char *ptr);
short PTPIP_Get_WORD(unsigned char *ptr);
void PTPIP_Store_WORD(unsigned char *ptr, short data);
void PTPIP_Store_DWORD(unsigned char *ptr, long data);
void PTPIP_pack_command(unsigned char *payload, int command, int length);
int PTPIP_Recv_Response(int sock, int *cmd, unsigned char *payload, int payload_len);
int PTPIP_Wait_Event(int sock, int wait, int *event_id, int event_param[]);

int ThetaShutter_UDPOpen(int port);
int ThetaShutter_UDPWaitTrigger(int sock, short *evshift);
int ThetaShutter_UDPClose(int sock);

int InitiateCapture(int nCommandSock, int nEventSock, int *nTransactionId, short evshift);

void AirPen_LED(int flag);



/* ----------------------------------------------
		Main
   ---------------------------------------------- */

int main(int argc, char *argv[]) {
	char sThetaAddr[256];
	int nThetaPort;
	int nMyPort;
	int nCommandSock;
	int nEventSock;
	int nTriggerSock;
	int nSessionId;
	int nTransactionId;
	int nRes;
	int aArgs[4];
	char *aData;
	int nDataLen;
	short nEvshift;

	if (argc != 4) {
		fprintf(stderr,"Usage : %s THETA_addr THETA_port Trigger_port\n",argv[0]);
		return 1;
	}

	/* parameters */
	sThetaAddr[255] = 0;
	strncpy(sThetaAddr,argv[1],254);
	nThetaPort	= atoi(argv[2]);
	nMyPort		= atoi(argv[3]);

	/* Wait port */
	nTriggerSock = ThetaShutter_UDPOpen(nMyPort);
	if (nTriggerSock<0) {
		fprintf(stderr,"ThetaShutter_UDPOpen Error\n");
		return 1;
	}

	/* Init_Command */
	nCommandSock = PTPIP_Open_Connection(sThetaAddr, nThetaPort);
	if (nCommandSock<0) {
		fprintf(stderr,"PTPIP_Open_Connection Error\n");
		return 1;
	}
	nRes = PTPIP_Init_Command_Request(nCommandSock);
	if (nRes<0) {
		fprintf(stderr,"PTPIP_Init_Command_Request Error\n");
		goto CommandSock_Close;
	}
	nRes = PTPIP_Wait_Init_Command_Ack(nCommandSock,&nSessionId);
	if (nRes<0) {
		fprintf(stderr,"PTPIP_Wait_Init_Command_Ack Error\n");
		goto CommandSock_Close;
	}

	/* Init_Event */
	nEventSock = PTPIP_Open_Connection(sThetaAddr, nThetaPort);
	if (nEventSock<0) {
		fprintf(stderr,"PTPIP_Open_Connection Error\n");
		return 1;
	}
	nRes = PTPIP_Init_Event_Request(nEventSock,nSessionId);
	if (nRes<0) {
		fprintf(stderr,"PTPIP_Init_Event_Request Error\n");
		goto EventSock_Close;
	}
	nRes = PTPIP_Wait_Init_Event_Ack(nEventSock);
	if (nRes<0) {
		fprintf(stderr,"PTPIP_Wait_Init_Event_Ack Error\n");
		goto EventSock_Close;
	}
	nTransactionId = 0;

	/* PTP_OC_OpenSession */
	aArgs[0] = nSessionId;
	nRes = PTPIP_Cmd_Request(nCommandSock,nTransactionId++, PTP_OC_OpenSession, aArgs, 1, NULL, 0);
	if (nRes<0) {
		fprintf(stderr,"PTPIP_Cmd_Request Error\n");
		goto EventSock_Close;
	}
	nRes = PTPIP_Wait_Cmd_Response(nCommandSock, &aData, &nDataLen);
	if (aData != NULL) { free(aData); }
	if (nRes<0) {
		fprintf(stderr,"PTPIP_Wait_Cmd_Response Error\n");
		goto EventSock_Close;
	}

	AirPen_LED(1);
	while(1) {
		/* Wait Trigger */
		nRes = ThetaShutter_UDPWaitTrigger(nTriggerSock, &nEvshift);
		if (nRes<0) {
			fprintf(stderr,"ThetaShutter_UDPWaitTrigger Error\n");
			goto EventSock_Close;
		}

		nRes = InitiateCapture(nCommandSock, nEventSock, &nTransactionId,  nEvshift);
		if (nRes<0) {
			fprintf(stderr,"InitiateCapture Error\n");
			goto EventSock_Close;
		}
	}


EventSock_Close:
	AirPen_LED(0);
	PTPIP_Close_Connection(nEventSock);
CommandSock_Close:
	PTPIP_Close_Connection(nCommandSock);
	ThetaShutter_UDPClose(nTriggerSock);

	return 0;
}

/* ----------------------------------------------
		Wait UDP Trigger
   ---------------------------------------------- */

int ThetaShutter_UDPOpen(int port) {
	int sock,res;
	struct sockaddr_in addr;

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock<0) { return -1; }
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = INADDR_ANY;
	res = bind(sock, (struct sockaddr *)&addr, sizeof(addr));
	if (res<0) { close(sock); return -1; }

	return sock;
}

int ThetaShutter_UDPWaitTrigger(int sock, short *evshift) {
	unsigned char buf[256];
	int res;

	*evshift = 0;
	while(1) {
		memset(buf,0, sizeof(buf));
		res = recv(sock, buf, sizeof(buf), 0);
		if (res<4) { return -1; }

		/* 'TRIG'+evshift(short) */
		if ((buf[0]=='T') && (buf[1]=='R') && (buf[2]=='I') && (buf[3]=='G')) {
			if (res<6) {
				*evshift = 0;
			} else {
				*evshift = PTPIP_Get_WORD(buf+4);
			}
			break;
		}
	}
	return 0;
}

int ThetaShutter_UDPClose(int sock) {
	close(sock);
	return 0;
}

/* ----------------------------------------------
		AirPen
   ---------------------------------------------- */

void AirPen_LED(int flag) {
	FILE *fp;
	fp = fopen(g_AirPenLED,"w");
	if (flag) {
		printf("Connect LED ON\n");
		if (fp != NULL) { fputc('1',fp); }
	} else {
		printf("Connect LED OFF\n");
		if (fp != NULL) { fputc('0',fp); }
	}
	if (fp != NULL) { fclose(fp); }
}


/* ----------------------------------------------
		Initiate Capture
   ---------------------------------------------- */
/* EV shift: 2000,1700,1300,1000,700,300,0,-300,-700,-1000,-1300,-1700,-2000 */

int InitiateCapture(int nCommandSock, int nEventSock, int *nTransactionId, short evshift) {
	int aArgs[4];
	int nRes;
	char aPayload[2];
	char *aData;
	int nDataLen;
	int nHandle;
	int loop;
	int event_id,event_param[3];

	/* LED */
	printf("InitiateCapture (%d)\n",evshift);
	AirPen_LED(0);

	/* PTP_OC_SetDevicePropValue */
	PTPIP_Store_WORD(aPayload, evshift);
	aArgs[0] = PTP_DPC_ExposureBiasCompensation;
	nRes = PTPIP_Cmd_Request(nCommandSock,*nTransactionId++, PTP_OC_SetDevicePropValue, aArgs, 1, aPayload, 2);
	if (nRes<0) {
		fprintf(stderr,"PTP_OC_SetDevicePropValue Error\n");
		return -1;
	}
	nRes = PTPIP_Wait_Cmd_Response(nCommandSock, &aData, &nDataLen);
	if (aData != NULL) { free(aData); }
	if (nRes<0) {
		fprintf(stderr,"PTPIP_Wait_Cmd_Response Error\n");
		return -1;
	}

	/* PTP_OC_InitiateCapture */
	aArgs[0] = 0;
	aArgs[1] = 0;
	nRes = PTPIP_Cmd_Request(nCommandSock,*nTransactionId++, PTP_OC_InitiateCapture, aArgs, 2, NULL, 0);
	if (nRes<0) {
		fprintf(stderr,"PTPIP_Cmd_Request Error\n");
		return -1;
	}
	nRes = PTPIP_Wait_Cmd_Response(nCommandSock, &aData, &nDataLen);
	if (aData != NULL) { free(aData); }
	if (nRes<0) {
		fprintf(stderr,"PTPIP_Wait_Cmd_Response Error\n");
		return -1;
	}

	/* Wait PTP_EC_CaptureComplete */
	nHandle = 0;
	for(loop=0;loop<20;loop++) {
		event_id = 0;
		nRes = PTPIP_Wait_Event(nEventSock, 500, &event_id, event_param);
		if (event_id == PTP_EC_CaptureComplete) {
			/* CaptureComplete */
			printf("CaptureComplete\n");
			break;
		} else if (event_id == PTP_EC_ObjectAdded) {
			/* Object Added */
			printf("Object Added\n");
			nHandle = event_param[0];
		}
	}

	/* PTP_OC_SetDevicePropValue */
	PTPIP_Store_WORD(aPayload, 0);
	aArgs[0] = PTP_DPC_ExposureBiasCompensation;
	nRes = PTPIP_Cmd_Request(nCommandSock,*nTransactionId++, PTP_OC_SetDevicePropValue, aArgs, 1, aPayload, 2);
	if (nRes<0) {
		fprintf(stderr,"PTP_OC_SetDevicePropValue Error\n");
		return -1;
	}
	nRes = PTPIP_Wait_Cmd_Response(nCommandSock, &aData, &nDataLen);
	if (aData != NULL) { free(aData); }
	if (nRes<0) {
		fprintf(stderr,"PTPIP_Wait_Cmd_Response Error\n");
		return -1;
	}

	/* LED */
	AirPen_LED(1);

	return nHandle;
}


/* ----------------------------------------------
		Init_Command
   ---------------------------------------------- */

int PTPIP_Init_Command_Request(int sock) {
	unsigned char aPayload[250];
	char *name;
	int i,len,res;

	/* GUID */
	for(i=0;i<16;i++) {
		aPayload[8+i] = g_sClientGUID[i];
	}
	/* Client Name */
	name = (char *)g_sClientName;
	i = 24;
	while(*name) {
		aPayload[i++] = *name++;
		aPayload[i++] = 0;
		if (i>=(sizeof(aPayload)-6)) {
			aPayload[i++] = 0;
			aPayload[i++] = 0;
			break;
		}
	}
	PTPIP_Store_DWORD(aPayload+i, 1); i+=4;

	len = i;
	PTPIP_pack_command(aPayload, 1, len);
	/* Send */
	res = write(sock, aPayload, len);
	if (res<0) { return -1; }
	return 0;
}

int PTPIP_Wait_Init_Command_Ack(int sock,int *session_id) {
	unsigned char aPayload[250];
	int	res,cmd;

	res = PTPIP_Recv_Response(sock, &cmd, aPayload, sizeof(aPayload));
	if (res < 0) { return -1; }
	if (cmd != 2) { return -1; }
	*session_id = PTPIP_Get_DWORD(aPayload);
	return 0;
}

/* ----------------------------------------------
		Init_Event
   ---------------------------------------------- */

int PTPIP_Init_Event_Request(int sock, int session_id) {
	unsigned char aPayload[16];
	int len,res;

	len = 12;
	PTPIP_Store_DWORD(aPayload+8, session_id);
	PTPIP_pack_command(aPayload, 3, len);
	/* Send */
	res = write(sock, aPayload, len);
	if (res<0) { return -1; }
	return 0;
}

int PTPIP_Wait_Init_Event_Ack(int sock) {
	unsigned char aPayload[250];
	int	res,cmd;

	res = PTPIP_Recv_Response(sock, &cmd, aPayload, sizeof(aPayload));
	if (res < 0) { return -1; }
	if (cmd != 4) { return -1; }
	return 0;
}


/* ----------------------------------------------
		Wait_Event
   ---------------------------------------------- */
/* event_param ... array[3] of int */

int PTPIP_Wait_Event(int sock, int wait, int *event_id, int event_param[]) {
	fd_set fds;
	struct timeval tv;
	int n;
	unsigned char aPayload[250];
	int	res,cmd,transaction_id,i;

	FD_ZERO(&fds);
	FD_SET(sock, &fds);
	tv.tv_usec = wait*1000;
	n = select(sock+1, &fds, NULL, NULL, &tv);
	if (n != 0) {
		res = PTPIP_Recv_Response(sock, &cmd, aPayload, sizeof(aPayload));
		if (res < 0) { return -1; }
		if (cmd != 8) { return -1; }

		*event_id      = PTPIP_Get_WORD(aPayload);
		transaction_id = PTPIP_Get_DWORD(aPayload+2);
		res -= 6;
		i   = 0;
		while((res>0) && (i<3)) {
			event_param[i] = PTPIP_Get_DWORD(aPayload+6+i*4);
			res -= 4;
			i++;
		}
		return 0;
	} else {
		return -1;
	}
}


/* ----------------------------------------------
		Cmd_Request
   ---------------------------------------------- */

int PTPIP_Cmd_Request(int sock, int transaction_id, int cmd, int arg[], int arglen, char *payload, int payloadlen) {
	unsigned char aPayload[250];
	int len,i,res,packetlen;

	len = PTPIP_PACK_START;
	PTPIP_Store_DWORD(aPayload+len, 1); len += 4;
	PTPIP_Store_WORD(aPayload+len, cmd); len += 2;
	PTPIP_Store_DWORD(aPayload+len, transaction_id); len += 4;
	for(i=0;i<arglen;i++) {
		PTPIP_Store_DWORD(aPayload+len, arg[i]); len += 4;
		if (len>=sizeof(aPayload)) {
			return -1;
		}
	}
	PTPIP_pack_command(aPayload, 6, len);
	/* Send */
	res = write(sock, aPayload, len);
	if (res<0) { return -1; }

	/* No payload */
	if ((payload == NULL) || (payloadlen==0)) { return 0; }

	/* Start Data Packet */
	len = PTPIP_PACK_START;
	PTPIP_Store_DWORD(aPayload+len, transaction_id); len += 4;
	PTPIP_Store_DWORD(aPayload+len, payloadlen); len += 4;
	PTPIP_Store_DWORD(aPayload+len, 0); len += 4;
	PTPIP_pack_command(aPayload, 9, len);
	/* Send */
	res = write(sock, aPayload, len);
	if (res<0) { return -1; }

	/* Data Packet */
	while(payloadlen>0) {
		packetlen = payloadlen;
		if (packetlen>200) { packetlen = 200; }

		len = PTPIP_PACK_START;
		PTPIP_Store_DWORD(aPayload+len, transaction_id); len += 4;
		memcpy(aPayload+len, payload, packetlen); len += packetlen;
		payloadlen -= packetlen;
		payload    += packetlen;
		if (payloadlen != 0) {
			/* Data Packet */
			PTPIP_pack_command(aPayload, 10, len);
		} else {
			/* End Data Packet */
			PTPIP_pack_command(aPayload, 12, len);
		}
		/* Send */
		res = write(sock, aPayload, len);
		if (res<0) { return -1; }
	}

	return 0;
}

int PTPIP_Wait_Cmd_Response(int sock, char **data, int *data_len) {
	unsigned char aPayload[250];
	int	res,cmd,result;
	int transaction_id, temp_transaction_id, data_payload_size, buffer_size;
	char *buffer,*buffer_ptr;

	buffer      = NULL;
	buffer_ptr  = NULL;
	buffer_size = 0;
	result      = -1;

	res = PTPIP_Recv_Response(sock, &cmd, aPayload, sizeof(aPayload));
	if (res < 0) { goto wait_cmd_response_end; }

	/* with Data? */
	if (cmd == 9) {
		/* Start Data Packet */
		transaction_id    = PTPIP_Get_DWORD(aPayload);
		data_payload_size = PTPIP_Get_DWORD(aPayload+4);
		buffer      = (char *)malloc(data_payload_size);
		if (buffer == NULL) { goto wait_cmd_response_end; }
		buffer_ptr  = buffer;
		buffer_size = data_payload_size;
		
		/* continuous */
		while(data_payload_size>0) {
			res = PTPIP_Recv_Response(sock, &cmd, aPayload, sizeof(aPayload));
			if (res < 0) { goto wait_cmd_response_end; }
			if ((cmd != 10) && (cmd != 12)) {
				/* not Data_Packet and End_Data_Packet */
				goto wait_cmd_response_end;
			}
			
			temp_transaction_id = PTPIP_Get_DWORD(aPayload);
			res = res - 4;	/* payload length */
			if (transaction_id != temp_transaction_id) { goto wait_cmd_response_end; }
			if ((res>data_payload_size) || (res<=0))   { goto wait_cmd_response_end; }
			memcpy(buffer_ptr, aPayload+4, res);
			buffer_ptr        += res;
			data_payload_size -= res;
			if (cmd == 12) { break; } /* End_Data_Packet */
		}

		/* cmd response */
		res = PTPIP_Recv_Response(sock, &cmd, aPayload, sizeof(aPayload));
		if (res < 0) { goto wait_cmd_response_end; }
	}
	
	/* Response? */
	if (cmd == 7) {
		/* Cmd Response */
		transaction_id    = PTPIP_Get_DWORD(aPayload);
		/* finish */
		result = 0;
	}
wait_cmd_response_end:
	if ((result<0) && (buffer != NULL)) {
		free(buffer);
		buffer = NULL;
	} else {
		*data     = buffer;
		*data_len = buffer_size;
	}

	return result;
}



/* ----------------------------------------------
		PTPIP utils
   ---------------------------------------------- */

int PTPIP_Open_Connection(char *addr, int port) {
	int sock;
	struct sockaddr_in server;
	struct hostent *host;

	memset(&server, 0, sizeof(server));
	server.sin_port		= htons(port);
	server.sin_family	= AF_INET;
	server.sin_addr.s_addr = inet_addr(addr);

	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock<0) { return -1; }

	if (server.sin_addr.s_addr == 0xffffffff) {
		host = gethostbyname(addr);
		if (host == NULL) {
			close(sock);
			return -1;
		}
		server.sin_addr.s_addr = *(unsigned int *)host->h_addr_list[0];
	}
	if (connect(sock, (struct sockaddr *)&server, sizeof(server))<0) {
		close(sock);
		return -1;
	}

	return sock;
}

int PTPIP_Close_Connection(int sock) {
	close(sock);
}

long PTPIP_Get_DWORD(unsigned char *ptr) {
	return (ptr[0] | (ptr[1] << 8) | (ptr[2] << 16) | (ptr[3] << 24));
}

short PTPIP_Get_WORD(unsigned char *ptr) {
	return (ptr[0] | (ptr[1] << 8));
}

void PTPIP_Store_WORD(unsigned char *ptr, short data) {
	ptr[ 0] = ( data        & 0xFF);
	ptr[ 1] = ((data >>  8) & 0xFF);
}

void PTPIP_Store_DWORD(unsigned char *ptr, long data) {
	ptr[ 0] = ( data        & 0xFF);
	ptr[ 1] = ((data >>  8) & 0xFF);
	ptr[ 2] = ((data >> 16) & 0xFF);
	ptr[ 3] = ((data >> 24) & 0xFF);
}

void PTPIP_pack_command(unsigned char *payload, int command, int length) {
	/* length */
	PTPIP_Store_DWORD(payload  , length);
	/* Command */
	PTPIP_Store_DWORD(payload+4, command);
}

int PTPIP_Recv_Response(int sock, int *cmd, unsigned char *payload, int payload_len) {
	int len,res;
	if (payload_len<4) { return -1; }
	/* packet length */
	res = read(sock, payload, 4);
	if (res<4) { return -1; }
	len = PTPIP_Get_DWORD(payload);
	/* command */
	res = read(sock, payload, 4);
	if (res<4) { return -1; }
	*cmd = PTPIP_Get_DWORD(payload);
	if (len==8) { return 0; }
	if (payload_len<(len-8)) { return -1; }
	res = read(sock, payload, (len-8));
	if (res<(len-8)) { return -1; }
	return (len-8);
}
